import 'dart:async';
import 'dart:ui';

import 'package:flutter/material.dart';
import 'package:flutter/scheduler.dart';
import 'package:rothko_mobile/rothko_system/snackbar/snackbar_route.dart';

import 'snackbar_route.dart' as route;

const String Snackbar_ROUTE_NAME = "/SnackbarRoute";

typedef void SnackbarStatusCallback(SnackbarStatus status);
typedef void OnTap(Snackbar Snackbar);

/// A highly customizable widget so you can notify your user when you fell like he needs a beautiful explanation.
class Snackbar<T> extends StatefulWidget {
  Snackbar(
      {Key? key,
      String? title,
      Color? textColor,
      String? message,
      Widget? titleText,
      Widget? messageText,
      Widget? icon,
      bool? shouldIconPulse = true,
      double? maxWidth,
      double? fontSize,
      EdgeInsets? margin = const EdgeInsets.all(0.0),
      EdgeInsets? padding = const EdgeInsets.all(16),
      double borderRadius = 0.0,
      Color? borderColor,
      double? borderWidth = 1.0,
      Color? backgroundColor = const Color(0xFF303030),
      Color? leftBarIndicatorColor,
      List<BoxShadow>? boxShadows,
      Gradient? backgroundGradient,
      Widget? trailingButton,
      OnTap? onTap,
      Duration? duration,
      bool? isDismissible = true,
      SnackbarDismissDirection dismissDirection =
          SnackbarDismissDirection.VERTICAL,
      bool? showProgressIndicator = false,
      AnimationController? progressIndicatorController,
      Color? progressIndicatorBackgroundColor,
      Animation<Color>? progressIndicatorValueColor,
      SnackbarPosition? snackbarPosition = SnackbarPosition.BOTTOM,
      SnackbarStyle? snackbarStyle = SnackbarStyle.FLOATING,
      Curve forwardAnimationCurve = Curves.easeOutCirc,
      Curve reverseAnimationCurve = Curves.easeOutCirc,
      Duration animationDuration = const Duration(seconds: 1),
      SnackbarStatusCallback? onStatusChanged,
      double? barBlur = 0.0,
      SnackbarRoute? snackbarRoute,
      bool? blockBackgroundInteraction = false,
      double? routeBlur,
      Color? routeColor,
      Form? userInputForm})
      : this.title = title!,
        this.message = message,
        this.titleText = titleText,
        this.messageText = messageText,
        this.icon = icon,
        this.shouldIconPulse = shouldIconPulse,
        this.maxWidth = maxWidth,
        this.margin = margin,
        this.padding = padding,
        this.fontSize = fontSize,
        this.onStatusChanged = onStatusChanged,
        this.borderRadius = borderRadius,
        this.borderColor = borderColor,
        this.borderWidth = borderWidth,
        this.backgroundColor = backgroundColor,
        this.leftBarIndicatorColor = leftBarIndicatorColor,
        this.boxShadows = boxShadows,
        this._snackbarRoute = snackbarRoute,
        this.textColor = textColor,
        this.backgroundGradient = backgroundGradient,
        this.mainButton = trailingButton,
        this.onTap = onTap,
        this.duration = duration,
        this.isDismissible = isDismissible,
        this.dismissDirection = dismissDirection,
        this.showProgressIndicator = showProgressIndicator,
        this.progressIndicatorController = progressIndicatorController,
        this.progressIndicatorBackgroundColor =
            progressIndicatorBackgroundColor,
        this.progressIndicatorValueColor = progressIndicatorValueColor,
        this.snackbarPosition = snackbarPosition,
        this.snackbarStyle = snackbarStyle,
        this.forwardAnimationCurve = forwardAnimationCurve,
        this.reverseAnimationCurve = reverseAnimationCurve,
        this.animationDuration = animationDuration,
        this.barBlur = barBlur,
        this.blockBackgroundInteraction = blockBackgroundInteraction,
        this.routeBlur = routeBlur,
        this.routeColor = routeColor,
        this.userInputForm = userInputForm,
        super(key: key) {
    this.onStatusChanged = onStatusChanged;
  }

  /// A callback for you to listen to the different Snackbar status
  SnackbarStatusCallback? onStatusChanged;

  SnackbarRoute? _snackbarRoute;

  /// The title displayed to the user
  final String? title;

  /// The message displayed to the user.
  final String? message;

  /// Replaces [title]. Although this accepts a [Widget], it is meant to receive [Text] or [RichText]
  final Widget? titleText;
  final Color? textColor;
  final double? fontSize;

  /// Replaces [message]. Although this accepts a [Widget], it is meant to receive [Text] or  [RichText]
  final Widget? messageText;

  /// Will be ignored if [backgroundGradient] is not null
  final Color? backgroundColor;

  /// If not null, shows a left vertical bar to better indicate the humor of the notification.
  /// It is not possible to use it with a [Form] and I do not recommend using it with [LinearProgressIndicator]
  final Color? leftBarIndicatorColor;

  /// [boxShadows] The shadows generated by Snackbar. Leave it null if you don't want a shadow.
  /// You can use more than one if you feel the need.
  /// Check (this example)[https://github.com/flutter/flutter/blob/master/packages/flutter/lib/src/material/shadows.dart]
  final List<BoxShadow>? boxShadows;

  /// Makes [backgroundColor] be ignored.
  final Gradient? backgroundGradient;

  /// You can use any widget here, but I recommend [Icon] or [Image] as indication of what kind
  /// of message you are displaying. Other widgets may break the layout
  final Widget? icon;

  /// An option to animate the icon (if present). Defaults to true.
  final bool? shouldIconPulse;

  /// Use if you need an action from the user. [FlatButton] is recommended here
  final Widget? mainButton;

  /// A callback that registers the user's click anywhere. An alternative to [mainButton]
  final OnTap? onTap;

  /// How long until Snackbar will hide itself (be dismissed). To make it indefinite, leave it null.
  final Duration? duration;

  /// True if you want to show a [LinearProgressIndicator].
  /// If [progressIndicatorController] is null, an infinite progress indicator will be shown
  final bool? showProgressIndicator;

  /// An optional [AnimationController] when you want to control the progress of your [LinearProgressIndicator].
  /// You are responsible for controlling the progress
  final AnimationController? progressIndicatorController;

  /// A [LinearProgressIndicator] configuration parameter.
  final Color? progressIndicatorBackgroundColor;

  /// A [LinearProgressIndicator] configuration parameter.
  final Animation<Color>? progressIndicatorValueColor;

  /// Determines if the user can swipe or click the overlay (if [routeBlur] > 0) to dismiss.
  /// It is recommended that you set [duration] != null if this is false.
  /// If the user swipes to dismiss or clicks the overlay, no value will be returned.
  final bool? isDismissible;

  /// Used to limit Snackbar width (usually on large screens)
  final double? maxWidth;

  /// Adds a custom margin to Snackbar
  final EdgeInsets? margin;

  /// Adds a custom padding to Snackbar
  /// The default follows material design guide line
  final EdgeInsets? padding;

  /// Adds a radius to all corners of Snackbar. Best combined with [margin].
  /// I do not recommend using it with [showProgressIndicator] or [leftBarIndicatorColor].
  final double? borderRadius;

  /// Adds a border to every side of Snackbar
  /// I do not recommend using it with [showProgressIndicator] or [leftBarIndicatorColor].
  final Color? borderColor;

  /// Changes the width of the border if [borderColor] is specified
  final double? borderWidth;

  /// Snackbar can be based on [SnackbarPosition.TOP] or on [SnackbarPosition.BOTTOM] of your screen.
  /// [SnackbarPosition.BOTTOM] is the default.
  final SnackbarPosition? snackbarPosition;

  /// [SnackbarDismissDirection.VERTICAL] by default.
  /// Can also be [SnackbarDismissDirection.HORIZONTAL] in which case both left and right dismiss are allowed.
  final SnackbarDismissDirection? dismissDirection;

  /// Snackbar can be floating or be grounded to the edge of the screen.
  /// If grounded, I do not recommend using a [backgroundColor] with transparency or [barBlur]
  final SnackbarStyle? snackbarStyle;

  /// The [Curve] animation used when show() is called. [Curves.easeOut] is default
  final Curve? forwardAnimationCurve;

  /// The [Curve] animation used when dismiss() is called. [Curves.fastOutSlowIn] is default
  final Curve? reverseAnimationCurve;

  /// Use it to speed up or slow down the animation duration
  final Duration? animationDuration;

  /// Default is 0.0. If different than 0.0, blurs only Snackbar's background.
  /// To take effect, make sure your [backgroundColor] has some opacity.
  /// The greater the value, the greater the blur.
  final double? barBlur;

  /// Determines if user can interact with the screen behind it
  /// If this is false, [routeBlur] and [routeColor] will be ignored
  final bool? blockBackgroundInteraction;

  /// Default is 0.0. If different than 0.0, creates a blurred
  /// overlay that prevents the user from interacting with the screen.
  /// The greater the value, the greater the blur.
  /// It does not take effect if [blockBackgroundInteraction] is false
  final double? routeBlur;

  /// Default is [Colors.transparent]. Only takes effect if [routeBlur] > 0.0.
  /// Make sure you use a color with transparency here e.g. Colors.grey[600].withOpacity(0.2).
  /// It does not take effect if [blockBackgroundInteraction] is false
  final Color? routeColor;

  /// A [TextFormField] in case you want a simple user input. Every other widget is ignored if this is not null.
  final Form? userInputForm;

  /// Show the Snackbar. Kicks in [SnackbarStatus.IS_APPEARING] state followed by [SnackbarStatus.SHOWING]
  Future<T> show(BuildContext context) async {
    _snackbarRoute = route.showSnackbar<T>(
      context: context,
      Snackbar: this,
    );

    return await Navigator.of(context, rootNavigator: false)
        .push(_snackbarRoute!);
  }

  /// Dismisses the Snackbar causing is to return a future containing [result].
  /// When this future finishes, it is guaranteed that Snackbar was dismissed.

  /// Checks if the Snackbar is visible
  bool isShowing() {
    return _snackbarRoute!.currentStatus == SnackbarStatus.SHOWING;
  }

  /// Checks if the Snackbar is dismissed
  bool isDismissed() {
    return _snackbarRoute!.currentStatus == SnackbarStatus.DISMISSED;
  }

  @override
  State createState() {
    return _SnackbarState<T>();
  }
}

class _SnackbarState<T> extends State<Snackbar> with TickerProviderStateMixin {
  final Duration _pulseAnimationDuration = const Duration(seconds: 1);
  final Widget _emptyWidget = const SizedBox();
  final double _initialOpacity = 1.0;
  final double _finalOpacity = 0.4;

  GlobalKey? _backgroundBoxKey;
  SnackbarStatus? currentStatus;
  AnimationController? _fadeController;
  Animation<double>? _fadeAnimation;
  bool? _isTitlePresent;
  double? _messageTopMargin;
  FocusScopeNode? _focusNode;
  FocusAttachment? _focusAttachment;
  Completer<Size>? _boxHeightCompleter;
  Function()? _progressListener;
  CurvedAnimation? _progressAnimation;

  @override
  void initState() {
    super.initState();

    _backgroundBoxKey = GlobalKey();
    _boxHeightCompleter = Completer<Size>();

    assert(((((widget.message != null) || widget.title != null))),
        "A message is mandatory if you are not using userInputForm. Set either a message or messageText");

    _isTitlePresent = (widget.title != null || widget.titleText != null);
    _messageTopMargin = _isTitlePresent! ? 0.0 : widget.padding!.top;

    _configureProgressIndicatorAnimation();

    if (widget.icon != null && widget.shouldIconPulse!) {
      _configurePulseAnimation();
      _fadeController!.forward();
    }

    _focusNode = FocusScopeNode();
    _focusAttachment = _focusNode!.attach(context);
  }

  @override
  void dispose() {
    _fadeController?.dispose();

    widget.progressIndicatorController?.removeListener(_progressListener!);
    widget.progressIndicatorController?.dispose();

    _focusAttachment!.detach();
    _focusNode!.dispose();
    super.dispose();
  }

  void _configureProgressIndicatorAnimation() {
    if (widget.showProgressIndicator! &&
        widget.progressIndicatorController != null) {
      _progressAnimation = CurvedAnimation(
          curve: Curves.linear, parent: widget.progressIndicatorController!);
    }
  }

  void _configurePulseAnimation() {
    _fadeController =
        AnimationController(vsync: this, duration: _pulseAnimationDuration);
    _fadeAnimation = Tween(begin: _initialOpacity, end: _finalOpacity).animate(
      CurvedAnimation(
        parent: _fadeController!,
        curve: Curves.linear,
      ),
    );

    _fadeController!.addStatusListener((status) {
      if (status == AnimationStatus.completed) {
        _fadeController!.reverse();
      }
      if (status == AnimationStatus.dismissed) {
        _fadeController!.forward();
      }
    });

    _fadeController!.forward();
  }

  @override
  Widget build(BuildContext context) {
    return Align(
      heightFactor: 1.0,
      child: Material(
        color: widget.snackbarStyle == SnackbarStyle.FLOATING
            ? Colors.transparent
            : widget.backgroundColor,
        child: SafeArea(
          minimum: widget.snackbarPosition == SnackbarPosition.BOTTOM
              ? EdgeInsets.only(
                  bottom: MediaQuery.of(context).viewInsets.bottom)
              : EdgeInsets.only(top: MediaQuery.of(context).viewInsets.top),
          bottom: widget.snackbarPosition == SnackbarPosition.BOTTOM,
          top: widget.snackbarPosition == SnackbarPosition.TOP,
          left: false,
          right: false,
          child: _getSnackbar(),
        ),
      ),
    );
  }

  Widget _getSnackbar() {
    Widget Snackbar;

    if (widget.userInputForm != null) {
      Snackbar = _generateInputSnackbar();
    } else {
      Snackbar = _generateSnackbar();
    }

    return Stack(
      children: [
        FutureBuilder(
          future: _boxHeightCompleter!.future,
          builder: (context, AsyncSnapshot<Size> snapshot) {
            if (snapshot.hasData) {
              return ClipRRect(
                borderRadius: BorderRadius.circular(widget.borderRadius!),
                child: BackdropFilter(
                  filter: ImageFilter.blur(
                      sigmaX: widget.barBlur!, sigmaY: widget.barBlur!),
                  child: Container(
                    height: snapshot.data!.height,
                    width: snapshot.data!.width,
                    decoration: BoxDecoration(
                      color: Colors.transparent,
                      borderRadius: BorderRadius.circular(widget.borderRadius!),
                    ),
                  ),
                ),
              );
            } else {
              return _emptyWidget;
            }
          },
        ),
        Snackbar,
      ],
    );
  }

  Widget _generateInputSnackbar() {
    return Container(
      key: _backgroundBoxKey,
      constraints: widget.maxWidth != null
          ? BoxConstraints(maxWidth: widget.maxWidth!)
          : null,
      decoration: BoxDecoration(
        color: widget.backgroundColor,
        gradient: widget.backgroundGradient,
        boxShadow: widget.boxShadows,
        borderRadius: BorderRadius.circular(widget.borderRadius!),
        border: widget.borderColor != null
            ? Border.all(color: widget.borderColor!, width: widget.borderWidth!)
            : null,
      ),
      child: Padding(
        padding: const EdgeInsets.only(
            left: 8.0, right: 8.0, bottom: 8.0, top: 16.0),
        child: FocusScope(
          child: widget.userInputForm!,
          node: _focusNode,
          autofocus: true,
        ),
      ),
    );
  }

  Widget _generateSnackbar() {
    return Container(
      key: _backgroundBoxKey,
      constraints: widget.maxWidth != null
          ? BoxConstraints(maxWidth: widget.maxWidth!)
          : null,
      decoration: BoxDecoration(
        color: widget.backgroundColor,
        gradient: widget.backgroundGradient,
        boxShadow: widget.boxShadows,
        borderRadius: BorderRadius.circular(widget.borderRadius!),
        border: widget.borderColor != null
            ? Border.all(color: widget.borderColor!, width: widget.borderWidth!)
            : null,
      ),
      child: Column(
        mainAxisSize: MainAxisSize.min,
        children: [
          Row(
            mainAxisSize: MainAxisSize.max,
            children: _getAppropriateRowLayout(),
          ),
        ],
      ),
    );
  }

  List<Widget> _getAppropriateRowLayout() {
    double buttonRightPadding;
    double iconPadding = 0;
    if (widget.padding!.right - 12 < 0) {
      buttonRightPadding = 4;
    } else {
      buttonRightPadding = widget.padding!.right - 12;
    }

    if (widget.padding!.left > 16.0) {
      iconPadding = widget.padding!.left;
    }

    if (widget.icon == null && widget.mainButton == null) {
      return [
        Expanded(
          flex: 1,
          child: Column(
            crossAxisAlignment: CrossAxisAlignment.stretch,
            mainAxisSize: MainAxisSize.min,
            children: <Widget>[
              (_isTitlePresent!)
                  ? Padding(
                      padding: EdgeInsets.only(
                        top: widget.padding!.top,
                        left: widget.padding!.left,
                        right: widget.padding!.right,
                      ),
                      child: _getTitleText(),
                    )
                  : _emptyWidget,
              Padding(
                padding: EdgeInsets.only(
                  top: _messageTopMargin!,
                  left: widget.padding!.left,
                  right: widget.padding!.right,
                  bottom: widget.padding!.bottom,
                ),
                child: widget.messageText,
              ),
            ],
          ),
        ),
      ];
    } else if (widget.icon != null && widget.mainButton == null) {
      return <Widget>[
        ConstrainedBox(
          constraints: BoxConstraints.tightFor(width: 32.0 + iconPadding),
          child: _getIcon(),
        ),
        Expanded(
          flex: 1,
          child: Column(
            crossAxisAlignment: CrossAxisAlignment.stretch,
            mainAxisSize: MainAxisSize.min,
            children: <Widget>[
              (_isTitlePresent!)
                  ? Padding(
                      padding: EdgeInsets.only(
                        top: widget.padding!.top,
                        left: 4.0,
                        right: widget.padding!.left,
                      ),
                      child: _getTitleText(),
                    )
                  : _emptyWidget,
              Padding(
                padding: EdgeInsets.only(
                  top: _messageTopMargin!,
                  left: 4.0,
                  right: widget.padding!.right,
                  bottom: widget.padding!.bottom,
                ),
                child: widget.messageText,
              ),
            ],
          ),
        ),
      ];
    } else if (widget.icon == null && widget.mainButton != null) {
      return <Widget>[
        Expanded(
          flex: 1,
          child: Column(
            crossAxisAlignment: CrossAxisAlignment.stretch,
            mainAxisSize: MainAxisSize.min,
            children: <Widget>[
              (_isTitlePresent!)
                  ? Padding(
                      padding: EdgeInsets.only(
                        top: widget.padding!.top,
                        left: widget.padding!.left,
                        right: widget.padding!.right,
                      ),
                      child: _getTitleText(),
                    )
                  : _emptyWidget,
              Padding(
                padding: EdgeInsets.only(
                  top: _messageTopMargin!,
                  left: widget.padding!.left,
                  right: 8.0,
                  bottom: widget.padding!.bottom,
                ),
                child: widget.messageText,
              ),
            ],
          ),
        ),
        Padding(
          padding: EdgeInsets.only(right: 20),
          child: _getMainActionButton(),
        ),
      ];
    } else {
      return <Widget>[
        Padding(
          padding: const EdgeInsets.only(left: 8.0),
          child: ConstrainedBox(
            constraints: BoxConstraints.tightFor(width: 22.0 + iconPadding),
            child: _getIcon(),
          ),
        ),
        Expanded(
          flex: 1,
          child: Column(
            crossAxisAlignment: CrossAxisAlignment.stretch,
            mainAxisSize: MainAxisSize.min,
            children: <Widget>[
              (_isTitlePresent!)
                  ? Padding(
                      padding: EdgeInsets.only(
                        top: widget.padding!.top,
                        left: 4.0,
                        right: 8.0,
                      ),
                      child: _getTitleText(),
                    )
                  : _emptyWidget,
              Padding(
                padding: EdgeInsets.only(
                  top: _messageTopMargin!,
                  left: 4.0,
                  right: 8.0,
                  bottom: widget.padding!.bottom,
                ),
                child: widget.messageText,
              ),
            ],
          ),
        ),
        Padding(
          padding: EdgeInsets.only(right: 20),
          child: _getMainActionButton(),
        )
      ];
    }
  }

  Widget? _getIcon() {
    if (widget.icon != null && widget.icon is Icon && widget.shouldIconPulse!) {
      return FadeTransition(
        opacity: _fadeAnimation!,
        child: widget.icon,
      );
    } else if (widget.icon != null) {
      return widget.icon!;
    } else {
      return _emptyWidget;
    }
  }

  Widget? _getTitleText() {
    return widget.titleText != null
        ? widget.titleText
        : Text(
            widget.title!,
            style: TextStyle(
                fontSize: widget.fontSize,
                color: widget.textColor,
                fontFamily: 'Proxima Nova',
                fontWeight: FontWeight.w400),
          );
  }

  Widget _getMainActionButton() {
    if (widget.mainButton != null) {
      return widget.mainButton!;
    } else {
      return SizedBox();
    }
  }
}

/// Indicates if Snackbar is going to start at the [TOP] or at the [BOTTOM]
enum SnackbarPosition { TOP, BOTTOM }

/// Indicates if Snackbar will be attached to the edge of the screen or not
enum SnackbarStyle { FLOATING }

/// Indicates the direction in which it is possible to dismiss
/// If vertical, dismiss up will be allowed if [SnackbarPosition.TOP]
/// If vertical, dismiss down will be allowed if [SnackbarPosition.BOTTOM]
enum SnackbarDismissDirection { HORIZONTAL, VERTICAL }

/// Indicates the animation status
/// [SnackbarStatus.SHOWING] SnackbarStatus has stopped and the user can see it
/// [SnackbarStatus.DISMISSED] Snackbar has finished its mission and returned any pending values
/// [SnackbarStatus.IS_APPEARING] Snackbar is moving towards [SnackbarStatus.SHOWING]
/// [SnackbarStatus.IS_HIDING] Snackbar is moving towards [] [SnackbarStatus.DISMISSED]
enum SnackbarStatus { SHOWING, DISMISSED, IS_APPEARING, IS_HIDING }
