import 'dart:math';

import 'package:flutter/material.dart';
import 'package:rothko_mobile/rothko.dart';

class DefaultToolTip extends StatelessWidget {
  final String? text;
  final FontWeight? fontWeight;

  final EdgeInsetsGeometry? padding;

  final double? borderRadius;

  final double? arrowWidth;

  final double? arrowHeight;

  final ArrowDirection? direction;

  final double? positionRatio;

  const DefaultToolTip({
    Key? key,
    this.padding,
    this.text,
    this.fontWeight,
    this.arrowWidth = 8.0,
    this.arrowHeight = 5.0,
    this.borderRadius = 10.0,
    this.direction = ArrowDirection.bottom,
    this.positionRatio = 0.5,
  }) : super(key: key);

  get _arrowMargin {
    var edgeInsets;
    switch (direction) {
      case ArrowDirection.left:
        edgeInsets = EdgeInsets.only(left: arrowHeight!);
        break;
      case ArrowDirection.top:
        edgeInsets = EdgeInsets.only(top: arrowHeight!);
        break;
      case ArrowDirection.right:
        edgeInsets = EdgeInsets.only(right: arrowHeight!);
        break;
      default:
        edgeInsets = EdgeInsets.only(bottom: arrowHeight!);
        break;
    }
    return edgeInsets;
  }

  @override
  Widget build(BuildContext context) {
    return Material(
        color: RothkoCache.getTheme!(context) == ThemeMode.light
            ? kLightNeutralBase
            : kDarkNeutralBase,
        shape: BubbleShape(
            direction: direction!,
            positionRatio: positionRatio!,
            arrowHeight: arrowHeight!,
            arrowWidth: arrowWidth!,
            borderRadius: borderRadius!),
        child: Container(
            margin: _arrowMargin,
            padding: padding,
            child: Text(
              text!,
              style: TextStyle(
                fontFamily: "Proxima",
                fontSize: 14,
                color: RothkoCache.getTheme!(context) == ThemeMode.light
                    ? kDarkNeutralBase
                    : kLightNeutralLighten0,
                fontWeight: fontWeight ?? FontWeight.w400,
              ),
            )));
  }
}

enum ArrowDirection { left, top, right, bottom }

enum BubbleStyle { stroke, fill }

class BubbleShape extends ShapeBorder {
  final BubbleStyle style;
  final Color strokeColor;
  final double strokeWidth;

  final ArrowDirection direction;

  final double arrowHeight;

  final double arrowWidth;

  final double borderRadius;
  final double positionRatio;

  BubbleShape(
      {this.style = BubbleStyle.fill,
      this.strokeColor = Colors.transparent,
      this.strokeWidth = 0.5,
      this.direction = ArrowDirection.bottom,
      this.positionRatio = 0.5,
      this.arrowHeight = 5.0,
      this.arrowWidth = 8.0,
      this.borderRadius = 10.0})
      : assert(positionRatio >= 0 && positionRatio <= 1, ''),
        assert(direction != null, '');

  getArrowHeightFit(Rect rect) {
    if (arrowHeight < 0) {
      return 0.0;
    }
    if (direction == ArrowDirection.left || direction == ArrowDirection.right) {
      if (arrowHeight > rect.width) {
        return rect.width;
      }
    } else {
      if (arrowHeight > rect.height) {
        return rect.height;
      }
    }
    return arrowHeight;
  }

  getBorderRadiusFit(Rect rect) {
    if (borderRadius < 0) {
      return 0.0;
    }
    var maxRadius;
    var arrowHeightFit = getArrowHeightFit(rect);
    if (direction == ArrowDirection.left || direction == ArrowDirection.right) {
      maxRadius = 0.5 * min(rect.width - arrowHeightFit, rect.height);
    } else {
      maxRadius = 0.5 * min(rect.width, rect.height - arrowHeightFit);
    }
    if (borderRadius > maxRadius) {
      return maxRadius;
    }
    return borderRadius;
  }

  getArrowWidthFit(Rect rect) {
    if (arrowWidth < 0) {
      return 0.0;
    }
    var borderRadiusFit = getBorderRadiusFit(rect);
    var maxWidth;
    if (direction == ArrowDirection.left || direction == ArrowDirection.right) {
      maxWidth = rect.height - 2 * borderRadiusFit;
    } else {
      maxWidth = rect.width - 2 * borderRadiusFit;
    }
    if (arrowWidth > maxWidth) {
      return maxWidth;
    }
    return arrowWidth;
  }

  getPositionRatioFit(Rect rect) {
    var borderRadiusFit = getBorderRadiusFit(rect);
    var arrowWidthFit = getArrowWidthFit(rect);
    var minPositionRatio;
    var maxPositionRatio;
    if (direction == ArrowDirection.left || direction == ArrowDirection.right) {
      minPositionRatio = (borderRadiusFit + 0.5 * arrowWidthFit) / rect.height;
      maxPositionRatio =
          (rect.height - borderRadiusFit - 0.5 * arrowWidthFit) / rect.height;
    } else {
      minPositionRatio = (borderRadiusFit + 0.5 * arrowWidthFit) / rect.width;
      maxPositionRatio =
          (rect.width - borderRadiusFit - 0.5 * arrowWidthFit) / rect.width;
    }
    if (positionRatio < minPositionRatio) {
      return minPositionRatio;
    }
    if (positionRatio > maxPositionRatio) {
      return maxPositionRatio;
    }
    return positionRatio;
  }

  @override
  Path getOuterPath(Rect rect, {TextDirection? textDirection}) {
    var path = Path();
    _addBubblePath(path, rect);
    return path;
  }

  @override
  void paint(Canvas canvas, Rect rect, {TextDirection? textDirection}) {
    if (style == BubbleStyle.stroke) {
      var paint = Paint()
        ..color = strokeColor
        ..strokeWidth = strokeWidth
        ..style = PaintingStyle.stroke
        ..strokeJoin = StrokeJoin.round;
      var path = Path();
      _addBubblePath(path, rect);
      canvas.drawPath(path, paint);
    }
  }

  

  _addBubblePath(Path path, Rect rect) {
    var w = rect.width;
    var h = rect.height;
    var borderRadiusFit = getBorderRadiusFit(rect);
    var positionRatioFit = getPositionRatioFit(rect);
    var arrowHeightFit = getArrowHeightFit(rect);
    var arrowWidthFit = getArrowWidthFit(rect);

    var xOffset = direction == ArrowDirection.left ? arrowHeightFit : 0.0;
    var yOffSet = direction == ArrowDirection.top ? arrowHeightFit : 0.0;
    var xOffsetEnd = direction == ArrowDirection.right ? arrowHeightFit : 0.0;
    var yOffSetEnd = direction == ArrowDirection.bottom ? arrowHeightFit : 0.0;

    var xArrowCenter = positionRatioFit * w;

    var yArrowCenter = positionRatioFit * h;
    path
      ..moveTo(xOffset, yOffSet + borderRadiusFit)
      ..arcTo(
          Rect.fromCircle(
              center:
                  Offset(xOffset + borderRadiusFit, yOffSet + borderRadiusFit),
              radius: borderRadiusFit),
          pi,
          0.5 * pi,
          false);

    if (direction == ArrowDirection.top) {
      path
        ..lineTo(xArrowCenter - 0.5 * arrowWidthFit, yOffSet)
        ..lineTo(xArrowCenter, 0.0)
        ..lineTo(xArrowCenter + 0.5 * arrowWidthFit, yOffSet);
    }
    path.lineTo(w - xOffsetEnd - borderRadiusFit, yOffSet);

    path.arcTo(
        Rect.fromCircle(
            center: Offset(
                w - xOffsetEnd - borderRadiusFit, yOffSet + borderRadiusFit),
            radius: borderRadiusFit),
        -0.5 * pi,
        0.5 * pi,
        false);

    if (direction == ArrowDirection.right) {
      path
        ..lineTo(w - xOffsetEnd, yArrowCenter - 0.5 * arrowWidthFit)
        ..lineTo(w, yArrowCenter)
        ..lineTo(w - xOffsetEnd, yArrowCenter + 0.5 * arrowWidthFit);
    }
    path.lineTo(w - xOffsetEnd, h - yOffSetEnd - borderRadiusFit);

    path.arcTo(
        Rect.fromCircle(
            center: Offset(w - xOffsetEnd - borderRadiusFit,
                h - yOffSetEnd - borderRadiusFit),
            radius: borderRadiusFit),
        0,
        0.5 * pi,
        false);

    if (direction == ArrowDirection.bottom) {
      path
        ..lineTo(xArrowCenter + 0.5 * arrowWidthFit, h - yOffSetEnd)
        ..lineTo(xArrowCenter, h)
        ..lineTo(xArrowCenter - 0.5 * arrowWidthFit, h - yOffSetEnd);
    }
    path.lineTo(xOffset + borderRadiusFit, h - yOffSetEnd);

    path.arcTo(
        Rect.fromCircle(
            center: Offset(
                xOffset + borderRadiusFit, h - yOffSetEnd - borderRadiusFit),
            radius: borderRadiusFit),
        0.5 * pi,
        0.5 * pi,
        false);

    if (direction == ArrowDirection.left) {
      path
        ..lineTo(xOffset, yArrowCenter + 0.5 * arrowWidthFit)
        ..lineTo(0.0, yArrowCenter)
        ..lineTo(xOffset, yArrowCenter - 0.5 * arrowWidthFit);
    }

    path.close();
  }

  @override
  EdgeInsetsGeometry get dimensions => throw UnimplementedError();

  @override
  Path getInnerPath(Rect rect, {TextDirection? textDirection}) {

      throw UnimplementedError();
    }
  
    @override
    ShapeBorder scale(double t) {

    throw UnimplementedError();
  }
}
